#!/usr/local/bin/python3.3
"""
Used to create sankey graphs showing the various
results for the papers analyzed in the study.
Usage:
sankey.py Total large graphics pdf
sankey.py Tissec large graphics pdf
"""
import numpy as np
import matplotlib.pyplot as plt
from pylab import *
from matplotlib.sankey import Sankey
import sys

##################################################################
from resultmacros import data
##################################################################

##################################################################
# Command line parsing
what = sys.argv[1]  # Total, NSF, Pldi, ...
kind = sys.argv[2]  # small or large
outDir = sys.argv[3]  # where to put output files
outType = sys.argv[4] # pdf, png, jpg, gif, ...

if kind == 'large':
   titleOn = True
   SCALE = 0.8
else:
   titleOn = False
   SCALE = 0.8

OutputFile = outDir + '/' + what + '.' + outType

##################################################################
# Customization
# http://matplotlib.org/users/customizing.html
# font.serif          : Bitstream Vera Serif, New Century Schoolbook, Century Schoolbook L, 
#                       Utopia, ITC Bookman, Bookman, Nimbus Roman No9 L, Times New Roman, 
#                       Times, Palatino, Charter, serif
# font.sans-serif     : Bitstream Vera Sans, Lucida Grande, Verdana, Geneva, Lucid, Arial, #
#                       Helvetica, Avant Garde, sans-serif
matplotlib.rcParams.update({
       'font.family': 'sans-serif',
       'font.size': 12, 
       'font.weight': 'bold'
    })

##################################################################

def title(t,w):
   if titleOn:
      return t + str(data[what + w])
   else:
      return str(data[what + w])

def titleSum(t,w1,w2):
   v = str(data[what + w1] + data[what + w2])
   if titleOn:
      return t + v
   else:
      return v

def titleSum4(t,w1,w2,w3,w4):
   v = str(data[what + w1] + data[what + w2] + data[what + w3] + data[what + w4])
   if titleOn:
      return t + v
   else:
      return v

# Add a fudge factor if a flow is 0, to ensre that the arrow
# shaft doesn't go away.
def fraction(w):
   v = data[what + w]
   if v == 0:
     return 0.005
   return v/data[what + 'Total']


def fractionSum(w1,w2):
   v = data[what + w1]+data[what + w2]
   if v == 0:
      return 0.005
   return v/data[what + 'Total']

def fractionSum4(w1,w2,w3,w4):
   v = data[what + w1] + data[what + w2] + data[what + w3] + data[what + w4]
   if v == 0:
      return 0.005
   return v/data[what + 'Total']

##################################################################

fig = plt.figure()
ax = fig.add_subplot(1, 1, 1, xticks=[], yticks=[], title="")
sankey = Sankey(ax=ax, 
                scale=SCALE, 
                offset=0.15, 
#                gap=0.2,
                head_angle=90,
                margin=0.0,
                format='%.0f', 
                unit=None)

########################################################################
# diagram 0 - input, and split out theoretical
# EQUATION:
#   Total = WithNoImplementation + WithImplementation + WithHardware
########################################################################
sankey.add(
   flows=[
      fraction('Total'),
         -fraction('WithNoImplementation'),
         -fraction('WithHardware'),
         -fraction('WithImplementation'),
    ],
    labels=[
       title('', 'Total'), 
       title('Theoretical\n','WithNoImplementation'), 
       title('Hard-\nware\n','WithHardware'), 
        '', 
    ],
    orientations=[
         0, 1, 1, 0
    ],
    pathlengths = [0.0, 0.3, 0.2, 0.0],
    linewidth=0.0,
    trunklength=1.0
)
########################################################################
# diagram 1 - where did we get code from?
# EQUATION:
#   WithImplementation = WithArticleCode + WithGoogleCode + WithEmailCode +
#                            (EmailWithoutReply + EmailGotReplyNo) +
#                            EmailNeededNotSent
########################################################################
sankey.add(
   flows=[
       fraction('WithImplementation'),
          -fraction('WithArticleCode'),
          -fraction('WithGoogleCode'),
          -fraction('WithEmailCode'),
          -fractionSum('EmailWithoutReply','EmailGotReplyNo'),
          -fraction('EmailNeededNotSent'),
    ],
    labels=[
       title('', 'WithImplementation'), 
       title('Article\n', 'WithArticleCode'), 
       title('Web\n', 'WithGoogleCode'), 
       title('Email\n', 'WithEmailCode'), 
       '',
       title('Email\nNot Sent\n', 'EmailNeededNotSent'), 
    ],
    orientations=[
         0, 1, 1, 1, 0, -1
    ],
   prior=0, 
   connect=(3, 0),
   trunklength = 0.5,
   linewidth=0.0,
   pathlengths = [0.1, 0.09, 0.21, 0.29, 0.1, 0.15]
)
########################################################################
# diagram 2 - intermediate step to merge all code sources
# EQUATION:
#   WithAvailableCode = WithArticleCode + WithGoogleCode + WithEmailCode
########################################################################
sankey.add(
   flows=[
          fraction('WithArticleCode'),
          fraction('WithGoogleCode'),
          fraction('WithEmailCode'),
          -fraction('WithAvailableCode'),
    ],
    labels=[
       '',
       '',
       '',
       ''
    ],
    orientations=[
         -1, -1, -1, 0
    ],
   prior=1, 
   connect=(1, 0),
   trunklength = 0.6,
   linewidth=0.0,
   pathlengths = [0.5, 0.5, 0.5, 0.0]
)
 
########################################################################
# diagram 3 - does it build?
# EQUATION:
#    WithAvailableCode = BuildDoesNotCompile +
#                            (BuildRuns + BuildOnlineRuns) +
#                            (BuildCompilesDoesNotRun + BuildOnlineDoesNotRun)
########################################################################
sankey.add(
   flows=[
       fraction('WithAvailableCode'),
           -fraction('BuildDoesNotCompile'),
           -fractionSum4('BuildRuns', 'BuildOnlineRuns','BuildCompilesDoesNotRun','BuildOnlineDoesNotRun'),
    ],
    labels=[
       title('', 'WithAvailableCode'),
       title('Build\nfails\n', 'BuildDoesNotCompile'), 
       titleSum4('Builds\n', 'BuildRuns', 'BuildOnlineRuns','BuildCompilesDoesNotRun','BuildOnlineDoesNotRun'), 
    ],
    orientations=[
        0, -1, 0
    ],
   prior=2, 
   connect=(3, 0),
   trunklength = 0.55,
   linewidth=0.0,
   pathlengths = [0.0, 0.1, 0.03]
)
########################################################################
# diagram 4 - does it run?
# EQUATION:
#     (BuildRuns + BuildOnlineRuns) + (BuildCompilesDoesNotRun + BuildOnlineDoesNotRun) =
#        (BuildCompilesDoesNotRun + BuildOnlineDoesNotRun) +
#        (BuildRuns + BuildOnlineRuns) 
########################################################################
#sankey.add(
#   flows=[
#       fractionSum4('BuildRuns', 'BuildOnlineRuns','BuildCompilesDoesNotRun','BuildOnlineDoesNotRun'),
#           -fractionSum('BuildCompilesDoesNotRun','BuildOnlineDoesNotRun'),
#           -fractionSum('BuildRuns','BuildOnlineRuns'),
#    ],
#    labels=[
#       titleSum4('', 'BuildRuns', 'BuildOnlineRuns','BuildCompilesDoesNotRun','BuildOnlineDoesNotRun'),
#       titleSum('Run\nfails\n', 'BuildCompilesDoesNotRun','BuildOnlineDoesNotRun'), 
#       titleSum('Runs\n', 'BuildOnlineRuns', 'BuildRuns'), 
#    ],
#    orientations=[
#        0, -1, 0
#    ],
#   prior=3, 
#   connect=(2, 0),
#   trunklength = 0.45,
#   linewidth=0.0,
#   pathlengths = [0.0, 0.1, 0.03]
#)

########################################################################
# diagram 5 - did they respond to emails?
# EQUATION:
#    EmailWithoutReply + EmailGotReplyNo =
#      EmailWithoutReply + EmailGotReplyNo
########################################################################
if ((data[what+'EmailWithoutReply']) + (data[what+'EmailGotReplyNo'])) > 0: 
   sankey.add(
      flows=[
          fractionSum('EmailWithoutReply', 'EmailGotReplyNo'),
             -fraction('EmailGotReplyNo'),
             -fraction('EmailWithoutReply'),
       ],
       labels=[
          titleSum('','EmailWithoutReply', 'EmailGotReplyNo'), 
          title('Email\nreply=no\n','EmailGotReplyNo'), 
          title('No email\nreply\n', 'EmailWithoutReply'), 
       ],
       orientations=[
            0, -1, -1
       ],
      prior=1, 
      connect=(4, 0),
      trunklength = 0.7,
      linewidth=0.0,
      pathlengths = [0.0, 0.3, 0.2]
   )

########################################################################
diagrams = sankey.finish()

diagrams[0].patch.set_facecolor('#80b060')
diagrams[1].patch.set_facecolor('#90c070')
diagrams[2].patch.set_facecolor('#90c070')
diagrams[3].patch.set_facecolor('#a0d080') 
#diagrams[4].patch.set_facecolor('#b0e090') 

if ((data[what+'EmailWithoutReply']) + (data[what+'EmailGotReplyNo'])) > 0: 
   diagrams[4].patch.set_facecolor('#a0d080') 
#   diagrams[5].patch.set_facecolor('#a0d080') 

diagrams[0].texts[1].set_color('b')
diagrams[0].texts[2].set_color('b')
#diagrams[1].texts[4].set_color('r')
diagrams[1].texts[5].set_color('orange')
#diagrams[1].texts[5].set_color('orange')
diagrams[3].texts[1].set_color('r')
diagrams[3].texts[2].set_color('g')
#diagrams[4].texts[1].set_color('r')
#diagrams[4].texts[2].set_color('g')

if ((data[what+'EmailWithoutReply']) + (data[what+'EmailGotReplyNo'])) > 0: 
   diagrams[4].texts[1].set_color('r')
   diagrams[4].texts[2].set_color('r')
#   diagrams[5].texts[1].set_color('r')
#   diagrams[5].texts[2].set_color('r')

fig.patch.set_visible(False)
ax.axis('off')
fig.tight_layout()
plt.savefig(filename=OutputFile,frameon=False) 
#plt.show()
